/* $Id: opt_reg.c,v 3.30 2001/09/09 05:16:38 jt Exp $ */
/* opt_reg.c */

/************************ register options ********************************
 *
 *    options are stored in an array,
 *    each element of which is a structure
 *
 *    The members of the structure contain information about
 *        1) the variable which is altered
 *        2) the type of variable
 *        3) a character by which the variable is invoked
 *        4) a string by which the variable can be invoked
 *        5) a brief description of the variable's role in the program
 *        6) a longer help message
 *        7) a flag indicating that an option has been invoked
 *        8) a flag indicating that the option is deprecated 
 *        9) a "hook", a function that is called when variable is invoked
 *       10) a pointer to an array structure if the option is for an array
 *            
 *    The structure elements are not assigned directly, but are
 *    "registered" with a call to the function optreg().
 *        
 *    In this file are the routines which access the structure of
 *    registered options.
 *
 *          I) Register options 
 *         II) Get number corresponding to name of option
 *        III) Print options out
 *             a) to a file which can be used as an input command line
 *             b) to the screen as part of a usage message
 *             c) to the screen as part of an interactive menu
 *         IV) Process single delimited and undelimited options
 */

#ifndef	SINGLEFILE
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "opt.h"
#include "ag.h"
#include "opt_p.h"
#endif

static void
opt_CSTRING_deprecated()
{
  opt_warning("OPT_CSTRING is deprecated...you should use OPT_STRING instead");
  opt_warning("If this warning message is a problem, then recompile the source");
  opt_warning("after removing the opt_CSTRING_deprecated() functions");
}

/* Default is to indicate whether a flag is true or false by using + or -
 * But, by #define'ing FLAG_ZERO_ONE, we will use 1 or 0 instead
 */
#ifdef FLAG_ZERO_ONE
#define FLAG_TRUE '1'
#define FLAG_FALSE '0'        
#else
#define FLAG_TRUE '+'
#define FLAG_FALSE '-'        
#endif

int opt_exit_number=0;          /* if we have to exit(), then exit
                                 * with this number.
                                 */

        /* -------------------------- */
        /* List of Registered Options */
        /* -------------------------- */


typedef struct {
  void **value;      /* pointer to array of data */
  int *size;         /* pointer to size of data */
  char delim;        /* char delimits elements of array: eg "," in "1,2,3" */
  char sdelim;       /* usually double quotes -- currently does nothing! */
  opt_TYPE base_type;  /* what the array of data is... */
}   OptArray_t;


typedef struct {
    void      *value;         /* pointer to value of option */
    opt_TYPE   type;          /* will be cast according to type */
    char       name;          /* name by which option is invoked */
    char      *longname;      /* long version of the name */
    char      *descript;      /* a brief description */
    char      *help;          /* longer help message */
    int        invoked;       /* flag is set to 1 if option is invoked */
    int        deprecated;    /* flag is set to 1 if option is deprecated
                                 added by Damian Eads (eads@lanl.gov)  */
    OPT_HOOK   hook;          /* code to evaluate if option is invoked */
    OptArray_t  *array;        /* set to null if not an array option */
}   Option_t;

static	Option_t *optlist=NULL;             /* array of options */
int              opt_nreg=0;                /* number of registered opts */

/* Prototypes for static (local) functions */
static void opt_fileheader(FILE *);
static int opt_action(int, ARGVECTOR *);
static int opt_number(void *);
static int opt_longname_number(char *);
static char *optarray_getstrval(int, void *, opt_TYPE);
static char *opt_getstrval( void *, opt_TYPE );
static int optsizeof(opt_TYPE);
static void opt_setvalue(void *, opt_TYPE, char *);
static void optarray_action(OptArray_t *, char *);

static char opt_array_delim=',';    /* the delimiter for arrays. (eads)*/
static char opt_array_string_delim='\"';

    /* ---------- */
    /* OPT macros */
    /* ---------- */

#define OPT_isvalidnum(n) ((n)>=0 && (n)<opt_nreg)

#define	OPT_isflagtype(o) \
	((o)==OPT_FLAG ||    (o)==OPT_NEGFLAG || \
         (o)==OPT_ABSFLAG || (o)==OPT_ABSNEGFLAG)
#define	OPT_isundelimtype(o) ((o)==OPT_UNDELIM ||  \
                              (o)==OPT_UNDELIMC || \
                              (o)==OPT_UNDELIMV )

/* OPT_GETVALUE: returns value of what 'void *' points to */
/* OPT_SETVALUE: sets the value that the 'void *' points to */

#define OPT_GETVALUE(typ,v)      ((typ)(*((typ *)(v))))
#define OPT_SETVALUE(typ,v,val)  do { typ *xptr;       \
                                      xptr = (typ *)v; \
                                      *xptr = (typ)val; } while (0)

/* OPT_GET_ITH_VALUE: gets value of i'th option */
/* OPT_SET_ITH_VALUE: sets value of i'th option */

#define	OPT_GET_ITH_VALUE(typ,i)     OPT_GETVALUE(typ,optlist[i].value)
#define OPT_SET_ITH_VALUE(typ,i,val) OPT_SETVALUE(typ,optlist[i].value,val)


        /* ------------------------------------------- */
        /* Routine to register each option as an array */
        /* Added by: Damian Eads <dre9227@cs.rit.edu>  */
        /* ------------------------------------------- */

static char *
optarray_getstrval( int size, void *val, opt_TYPE otype )
{
  int n;
  int len=0;
  char *str_val = NULL; /* string with all values in array 
                          * will be re-allocated for each item
                          */
  /* Convert the array 'val' into a string 'str_val' */
  for ( n = 0; n < size; n++ ) {
    /* this next line may look a little arcane...
     * 1. Recall val is in real life a pointer to an array, that is type **
     *    but we don't know the type, so we pass it through the argument list
     *    as a void *
     * 2. *(void **)val recasts val to a void ** and then dereferences it
     *    so that it is the pointer that is the array, but its type is
     *    `void *' instead of `type *'
     * 3. (char *)(*(void **)val) recasts this generic `void *' to be
     *    a `char *' because we know what the size of a char is (it's 1)
     *    and we can do arithmetic with it
     * 4. n*optsizeof(otype) is the offset of the n'th element of the array
     * 5. nth_val points to the n'th value in the array and if we were to
     *    recast it as *(type *)nth_val it would be the explicit value.
     *    Of course we don't know type, but we do know otype, and so
     *    we let opt_getstrval() take care of that
     */
    char *nth_val = (char *)(*(void **)val) + n*optsizeof( otype );
    char *str_nth_val = opt_getstrval( nth_val, otype );
    char str_delim[2] = { opt_array_delim, '\0' }; 

    if ( n == size-1 ) str_delim[0]='\0';  /* no delimiter at the end */
      
    /* Make room in str_val for the next item */
    len += strlen(str_nth_val) + strlen(str_delim);
    if (str_val) {
      str_val = (char *)realloc(str_val,sizeof(char)*(len+1));
    } else {
      str_val = (char *)malloc(sizeof(char)*(len+1));
      str_val[0]='\0';
    }
    strcat(str_val,str_nth_val);
    strcat(str_val,str_delim);
  }

  return str_val;
}

int
optrega_array(  int *size, void *val, opt_TYPE otype, char name,
                char *longname, char *descript )
{
  static char **str_val_array=NULL;
  static int    str_val_array_n=0;

    ++str_val_array_n;
    str_val_array = (char **)realloc(str_val_array,
                                     str_val_array_n*sizeof(char *));
    str_val_array[str_val_array_n-1] = optarray_getstrval(*size,val,otype);
  
    /* Register the option as an OPT_STRING */
    optrega( &str_val_array[str_val_array_n-1],
             OPT_STRING, name, longname, descript );

    /* Warn if inappropriate base_type is specified */
    switch (otype) {
      /* These are all okay */
    case OPT_INT:
    case OPT_UINT:
    case OPT_UNSINT:
    case OPT_SHORT:
    case OPT_USHORT:
    case OPT_LONG:
    case OPT_ULONG:
    case OPT_FLOAT:
    case OPT_DOUBLE:
      break;

      /* These _should_ be okay too */
    case OPT_STRING:
    case OPT_VSTRING:
      break;

      /* These might be problematic */
    case OPT_CSTRING:    
    case OPT_UNDELIMC:
      opt_CSTRING_deprecated(); /* no break */
    case OPT_CHAR:
    case OPT_UCHAR:
    case OPT_INTLEVEL:
    case OPT_FLAG:
    case OPT_ABSFLAG:
    case OPT_NEGFLAG:
    case OPT_ABSNEGFLAG:
    case OPT_UNDELIM:
    case OPT_UNDELIMV:
    case OPT_NUL:
      opt_warn_1("Undefined behavior for %s array",optstrtyp(opt_nreg-1));
      break;

    default:
        opt_fatal("opt_getstrval: Undefined o-type");
        break;
    }

    /* Allocate an OptArray_t structure */
    optlist[ opt_nreg-1 ].array = (OptArray_t *)malloc(sizeof(OptArray_t));
                                   
    /* Then add array-specific information to the registered option */
    optlist[ opt_nreg-1 ].array->value     = (void **)val;
    optlist[ opt_nreg-1 ].array->size      = size;
    optlist[ opt_nreg-1 ].array->base_type = otype;
    optlist[ opt_nreg-1 ].array->delim     = opt_array_delim;
    optlist[ opt_nreg-1 ].array->sdelim    = opt_array_string_delim;

    
    return opt_nreg-1;
}

int optreg_array( int *size, void *val, opt_TYPE otype, char name,
		  char *descript ) {
    return optrega_array(size,val,otype,name,NULL,descript);
}
int optregc_array( int *size, void *val, opt_TYPE otype, char name ) {
    return optrega_array(size,val,otype,name,NULL,NULL);
}
int optregcb_array( int *size, void *val, opt_TYPE otype, char name,
		    char *descript ) {
    return optrega_array(size,val,otype,name,NULL,descript);
}
int optregs_array( int *size, void *val, opt_TYPE otype, char *longname ) {
    return optrega_array(size,val,otype,'\0',longname,longname);
}
int optregsb_array( int *size, void *val, opt_TYPE otype, char *longname,
		    char *descript) {
    return optrega_array(size,val,otype,'\0',longname,descript);
}


        /* ------------------------------- */
        /* Routine to register each option */
        /* ------------------------------- */

int optreg(void *val, opt_TYPE otype, char name, char *descript) {
    return optrega(val,otype,name,NULL,descript);
}
int optregc(void *val, opt_TYPE otype, char name) {
    return optrega(val,otype,name,NULL,NULL);
}
int optregcb(void *val, opt_TYPE otype, char name, char *descript) {
    return optrega(val,otype,name,NULL,descript);
}
int optregs(void *val, opt_TYPE otype, char *longname) {
    return optrega(val,otype,'\0',longname,longname);
}
int optregsb(void *val, opt_TYPE otype, char *longname, char *descript) {
    return optrega(val,otype,'\0',longname,descript);
}
int optexec(char *longname, OPT_HOOK fcn, char *descript)
{
    int n;
    n = optrega(NULL,OPT_NUL,'\0',longname,descript);
    opthook_n(n,fcn);
    return n;
}
     
int
optrega(void *val, opt_TYPE otype, char name, char *longname, char *descript)
{
    /* write a warning if that option name has been used before */
    if( name != '\0' && opt_char_number(name) != -1 ) {
        opt_warn_1("optrega: Duplicate option name \'%c\'",name);
    }
    if (!ISEMPTYSTRING(longname) && opt_longname_number(longname) != -1) {
        opt_warn_1("optrega: Duplicate option longname \"%s\"",longname);
    }

    /* val points to the memory where the value of the option
     * is maintained.  This cannot be NULL.
     */
    if( val == NULL && otype != OPT_NUL ) {
        if (name)     { opt_warn_1("Name: -%c", name); }
        if (longname) { opt_warn_1("Name: --%s",longname); }
        opt_fatal("optrega: invalid first argument=NULL\n");
    }
    
    /* This is probably a mistake to print out a warning message,
       since it will be printed every time the program is run.
       Perhaps it should only be run if the user types 'run --testopt'
       */
#if 0    
    if( val != NULL  && opt_val_number(val) != -1) {
        opt_warn_1("optrega: This variable [%s] has already been registered",
                   thename);
    }
#endif    
    
    /* Programmer may specify that an option is to be an
     * undelimted option by setting the name to '\0'
     */
    if(name=='\0' && ISEMPTYSTRING(longname)) {
        if(otype==OPT_STRING)  otype=OPT_UNDELIM;
        if(otype==OPT_VSTRING) otype=OPT_UNDELIM;
        if(otype==OPT_CSTRING) otype=OPT_UNDELIMC;
    }

    /* Make sure that the longname doesn't have any whitespace */
    if (!ISEMPTYSTRING(longname)) {
     	char *t = longname;
        while (*t != '\0') {
          if (isspace(*t)) {
            opt_fatal_1("invalid longname [%s] contains whitespace\n",longname);
          }
          ++t;
        }
    }

    /* reallocate another struct for the new opt */
    ++opt_nreg;
    optlist = (Option_t *)realloc(optlist,opt_nreg*sizeof(Option_t));

    /* Having checked for various warnings, now register the options */
    /* FIXME: May want to strdup the longname and descript in case those
       names are set from the same variable set to different values */
    
    optlist[opt_nreg-1].value         = val;
    optlist[opt_nreg-1].type          = otype;
    optlist[opt_nreg-1].name          = name;
    optlist[opt_nreg-1].longname      = longname;
    optlist[opt_nreg-1].descript      = descript;

    optlist[opt_nreg-1].invoked       = 0;
    optlist[opt_nreg-1].deprecated    = 0;
    optlist[opt_nreg-1].help          = NULL;
    optlist[opt_nreg-1].hook          = NULL;
    optlist[opt_nreg-1].array         = NULL;

    /* UN-delimited strings are set to NULL when registered */
    /* This enables them to be invoked on the command line  */
    
    if (val != NULL) {
        if(otype==OPT_UNDELIMC) *((char *)optlist[opt_nreg-1].value)='\0';
        if(otype==OPT_UNDELIM ||
           otype==OPT_UNDELIMV) OPT_SET_ITH_VALUE(char *,opt_nreg-1,NULL);
    }
    return opt_nreg-1;

}
void optchar_n(int n, char name)
{
    if (OPT_isvalidnum(n)) optlist[n].name = name;
}    
void optdescript_n(int n, char *descript)
{
    if (OPT_isvalidnum(n)) optlist[n].descript = strdup(descript);
}
void opthelp_n(int n, char *help)
{
    if (OPT_isvalidnum(n)) optlist[n].help = strdup(help);
}
void optlongname_n(int n, char *longname)
{
    if (OPT_isvalidnum(n)) optlist[n].longname = strdup(longname);
}
void opthook_n(int n, OPT_HOOK hook)
{
    if (OPT_isvalidnum(n)) optlist[n].hook = hook;
}
int optinvoked_n(int n)
{
    if (OPT_isvalidnum(n))
        return optlist[n].invoked;
    return 0;
}

int optinvoked(void *v) { return optinvoked_n( opt_number(v) ); }

void optchar(void *v,  char name) { optchar_n( opt_number(v),   name ); }
void opthelp(void *v,  char *help) { opthelp_n( opt_number(v),  help ); }
void opthook(void *v,  OPT_HOOK hook) { opthook_n( opt_number(v), hook ); }      
void optlongname(void *v, char *longname)
{
    optlongname_n( opt_number(v), longname );
}
void optdescript(void *v, char *descript)
{
    optdescript_n( opt_number(v), descript );
}
void optExitNumber(int n) {
    opt_exit_number = n;
}


void
opt_get_help(char c)
{
    int n;
    n = opt_char_number(c);
    if (OPT_isvalidnum(n) && optlist[n].help != NULL)
        opt_mess_2("%c: %s\n",c,optlist[n].help);
    else
        opt_mess_1("Help unavailable for \'%c\'\n",c);
}
	
	

/* opt_XXX_number(): Get number corresponding to option name; return
 * a value of -1 if the option does not exist.
 *  opt_char_number(c) takes as input the single-character name;
 *  opt_longname_number(longname) takes as input the long name
 *  opt_number(void *) takes as input pointer to variable
 */

int
opt_char_number(char c)
{
    int i;                         /* see which registered option */
    for(i=0; i<opt_nreg; ++i) {
        if( c == optlist[i].name )
            return i;
    }
    return -1;
}
    
static int
opt_number(void *v)
{
    int i;                         /* see which registered option */
    for(i=0; i<opt_nreg; ++i) {
        if( v == optlist[i].value )
            return i;
    }
    return -1;                     /* to signify not an option */
}
static int
opt_longname_number(char *s)
{
    int i;                         /* see which registered option */
    if ( ISEMPTYSTRING(s) ) return -1;
    for(i=0; i<opt_nreg; ++i) {
        if ( !ISEMPTYSTRING(optlist[i].longname) ) {
            if (0==strcmp(s,optlist[i].longname)) {
                return i;
            }
        }
    }
    return -1;                     /* to signify not an option */
}

	/* -------------------------------------------- */
	/* Print value of registered option to a string */
	/* -------------------------------------------- */

/* optstrval:
 * returns a string corresponding to the ith option
 * Thus if x = 12.6, this returns the string "12.6"
 * Calls opt_getstrval for all types except INTLEVEL.
 * The return value is a pointer to a static string which is
 * overwritten with each call
 * FLAG values are returned as "+" or "-"
 * INTLEVEL values are returned as "- -xxxx..." depending on value
 */

char *
optstrval(int i)
{
    static char stval_buf[80];

    switch( optlist[i].type ) {
    case OPT_INTLEVEL:
        if (1) {
            int j;
            char buff[80];
            strcpy(stval_buf,"-"); 		/* Begin with level==OFF */
            for(j=0; j< OPT_GET_ITH_VALUE(int,i); ++j) {
                sprintf(buff,(j==0?" -%c":"%c"),optlist[i].name);
                strcat(stval_buf,buff);
            }
        }
        return stval_buf;
        break;
    default:
      return opt_getstrval(optlist[i].value,optlist[i].type);
      break;
    }
}/*optstrval*/

	/* -------------------------------------------- */
	/* Print value of registered option to a string */
        /* without specifying an index. (eads)          */
	/* -------------------------------------------- */

static char *
opt_getstrval( void *val, opt_TYPE typ ) {
    int maybe;
    static char stval_buf[80];

    switch( typ ) {
    case OPT_INT:
        sprintf(stval_buf,"%d", OPT_GETVALUE(int,val) );
        break;
    case OPT_UINT:
    case OPT_UNSINT:
        sprintf(stval_buf,"%u", OPT_GETVALUE(unsigned int,val) );
        break;
    case OPT_SHORT:
        sprintf(stval_buf,"%d", OPT_GETVALUE(short,val) );
        break;
    case OPT_USHORT:
        sprintf(stval_buf,"%u", OPT_GETVALUE(unsigned short,val) );
        break;
    case OPT_LONG:
        sprintf(stval_buf,"%ld", OPT_GETVALUE(long,val) );
        break;
    case OPT_ULONG:
        sprintf(stval_buf,"%lu", OPT_GETVALUE(unsigned long,val) );
        break;
    case OPT_CHAR:
        sprintf(stval_buf,"%c", OPT_GETVALUE(char,val));
        break;
    case OPT_UCHAR:
        sprintf(stval_buf,"%c", OPT_GETVALUE(unsigned char,val));
        break;
    case OPT_FLOAT:
        sprintf(stval_buf,"%g", OPT_GETVALUE(float,val) );
        break;
    case OPT_DOUBLE:
        sprintf(stval_buf,"%g", OPT_GETVALUE(double,val) );
        break;
    case OPT_INTLEVEL:
        opt_warning("Cannot use standard '-v- -vvvv' style setting for OPT_INTLEVEL");
        if (1) {
          int v = OPT_GETVALUE(int,val);
          if (v < 0) {
            opt_fatal("Invalid [negative] value for OPT_INTLEVEL");
          }
          if (v < 10) {
            sprintf(stval_buf,"%1d",v);
          }
          if (v >= 10) {
            opt_warning("Cannot set string for OPT_INTLEVEL");
          }
        }
        break;
    case OPT_FLAG:
    case OPT_ABSFLAG:
        maybe = OPT_GETVALUE(int,val);
        sprintf(stval_buf,"%c", (maybe==OPT_TRUE ? FLAG_TRUE : FLAG_FALSE) );
        break;
    case OPT_NEGFLAG:
    case OPT_ABSNEGFLAG:
        maybe = OPT_GETVALUE(int,val);
        sprintf(stval_buf,"%c", (maybe==OPT_TRUE ? FLAG_FALSE : FLAG_TRUE) );
        break;
    case OPT_CSTRING:    
    case OPT_UNDELIMC:
        opt_CSTRING_deprecated();
        sprintf(stval_buf,"\"%s\"",(char *)val);
        break;
    case OPT_STRING:
    case OPT_VSTRING:
    case OPT_UNDELIM:
    case OPT_UNDELIMV:
        if( val == NULL )	/* this should never happen */
            sprintf(stval_buf,"\"\"");
        else 
            if( OPT_GETVALUE(char *,val) == NULL )
                sprintf(stval_buf,"\"\"");
            else
                sprintf(stval_buf,"\"%s\"",OPT_GETVALUE(char *,val));
        break;
    case OPT_NUL:
        break;
    default:
        opt_fatal("opt_getstrval: Undefined o-type");
        break;
    }
    /* ---- Return value is static buffer ---- */
    return stval_buf;
}/* opt_getstrval */

static int
optsizeof( opt_TYPE typ )
{
    int retval;
    
    switch( typ ) {
    case OPT_INT:
        retval = sizeof(int);
        break;
    case OPT_UINT:
    case OPT_UNSINT:
        retval = sizeof(unsigned int);
        break;
    case OPT_SHORT:
        retval = sizeof(short);
        break;
    case OPT_USHORT:
        retval = sizeof(unsigned short);
        break;
    case OPT_LONG:
        retval = sizeof(long);
        break;
    case OPT_ULONG:
        retval = sizeof(unsigned long);
        break;
    case OPT_CHAR:
        retval = sizeof(char);
        break;
    case OPT_UCHAR:
        retval = sizeof(unsigned char);
        break;
    case OPT_FLOAT:
        retval = sizeof(float);
        break;
    case OPT_DOUBLE:
        retval = sizeof(double);
        break;
    case OPT_INTLEVEL:
      // FIXME
        retval = sizeof(double);
        break;
    case OPT_FLAG:
    case OPT_ABSFLAG:
        retval = sizeof(int);
        break;
    case OPT_NEGFLAG:
    case OPT_ABSNEGFLAG:
        retval = sizeof(int);
        break;
    case OPT_CSTRING:
    case OPT_UNDELIMC:
        opt_CSTRING_deprecated();
        retval = sizeof(char*);
        break;
    case OPT_STRING:
    case OPT_VSTRING:
    case OPT_UNDELIM:
    case OPT_UNDELIMV:
        retval = sizeof(char*);
        break;
    case OPT_NUL:
      // FIXMe
        retval = sizeof(int(*)());
        break;
    default:
        opt_fatal("optsizeof: Undefined o-type");
        break;
    }
    return retval;
}/*optsizeof*/

/* optstrtyp:
 * returns the type of the ith registered option as a string
 * Thus if x = 12.6 is a float , this returns the string "<FLOAT>"
 * The return value is a pointer to a static string which is
 * overwritten with each call
 */
char *
optstrtyp(int i)
{
    static char	sttyp_buf[80];
    opt_TYPE o;

    o = (optlist[i].array) ? optlist[i].array->base_type : optlist[i].type;
    
    switch( o ) {
        /* Integers */
    case OPT_INT:         strcpy(sttyp_buf,"<INT"); break;
    case OPT_UNSINT:      /* deprecated, same as UINT */
    case OPT_UINT:        strcpy(sttyp_buf,"<UINT"); break;
    case OPT_SHORT:       strcpy(sttyp_buf,"<SHORT"); break;
    case OPT_LONG:        strcpy(sttyp_buf,"<LONG"); break;
    case OPT_USHORT:      strcpy(sttyp_buf,"<USHORT"); break;
    case OPT_ULONG:       strcpy(sttyp_buf,"<ULONG"); break;
        /* Floating point */
    case OPT_FLOAT:       strcpy(sttyp_buf,"<FLOAT"); break;
    case OPT_DOUBLE:      strcpy(sttyp_buf,"<DOUBLE"); break;
        /* Char */
    case OPT_CHAR:        strcpy(sttyp_buf,"<CHAR"); break;
    case OPT_UCHAR:       strcpy(sttyp_buf,"<UCHAR"); break;
        /* Flag */
    case OPT_INTLEVEL:    strcpy(sttyp_buf,"<INTLEVEL"); break;
    case OPT_FLAG:        strcpy(sttyp_buf,"<FLAG"); break;
    case OPT_NEGFLAG:     strcpy(sttyp_buf,"<NEGFLAG"); break;
    case OPT_ABSFLAG:     strcpy(sttyp_buf,"<ABSFLAG"); break;
    case OPT_ABSNEGFLAG:  strcpy(sttyp_buf,"<ABSNEGFLAG"); break;
        /* String */
    case OPT_STRING:      strcpy(sttyp_buf,"<STRING"); break;
    case OPT_VSTRING:     strcpy(sttyp_buf,"<STRING"); break;
    case OPT_CSTRING:     strcpy(sttyp_buf,"<STRING"); break;
        /* Undelimited */
    case OPT_UNDELIM:     strcpy(sttyp_buf,"<UNDELIM"); break;
    case OPT_UNDELIMV:    strcpy(sttyp_buf,"<UNDELIM"); break;
    case OPT_UNDELIMC:    strcpy(sttyp_buf,"<UNDELIM"); break;
        /* Nul */
    case OPT_NUL:         strcpy(sttyp_buf,"<NUL"); break;
        
    default:
        opt_fatal("usage: undefined o-type");
    }
    /* If it's an array, then add " ARRAY" to the name of the type */
    if ( optlist[i].array ) {
      strcat(sttyp_buf," ARRAY");
    }
    /* finish with closing bracket */
    strcat(sttyp_buf, ">");
    return sttyp_buf;
}

		/* ------------------------------ */
		/* Put registered options to file */
		/* ------------------------------ */

void
opt_tofile(FILE *fp)
{
    int	i;
    char *format;
    /* Short Option Names */
    char *fnumstr="-%c %-38s %c%s\n";	/* numbers and strings  */
    char *fflgchr="-%c%-38s  %c%s\n";	/* flags and characters */
    
    opt_fileheader(fp);
    /* Short Option Names First */
    for(i=0; i<opt_nreg; ++i) {
        if (optlist[i].name == '\0')
            continue;
        if( OPT_isundelimtype( optlist[i].type ) )
            continue;
        if (optlist[i].type == OPT_NUL)
            continue;
        if( OPT_isflagtype( optlist[i].type )
            || optlist[i].type == OPT_INTLEVEL
            || optlist[i].type == OPT_CHAR )
            format = fflgchr;
        else
            format = fnumstr;
        fprintf(fp,format,optlist[i].name,
                optstrval(i),IGNOREEOL,optlist[i].descript);
        /* If an option has both a short and a long option name
         * then write it as a comment '    ; -b == --option-b
         */
        if (optlist[i].longname) {
            fprintf(fp,"%-42s%c -%c == --%s\n","",IGNOREEOL,
                    optlist[i].name,optlist[i].longname);
        }
    }
    /* Now Long Option Names (thanks to Jason V. Morgan) */
    for(i=0; i<opt_nreg; ++i) {
        /* do not output the long name if a short name exists */
        if (optlist[i].name != '\0')
           continue;
        if (!optlist[i].longname) 
           continue;

        /* The following arguably applies to short options too ?? */
        /* don't want anything with a hook (do we?) */
        if (optlist[i].type == OPT_NUL)
            continue;
        if (optlist[i].hook != NULL)
            continue;
        
        fprintf(fp, "--%s=%-*s %c%s\n", optlist[i].longname, 
                38 - strlen(optlist[i].longname), optstrval(i), 
                IGNOREEOL, optlist[i].descript);
    }
}
static void
opt_fileheader(FILE *fp)
{
    extern char *opt_program_name;
    int	i,n;
    char buf[80];

    sprintf(buf,"Options file created by: [%.30s]",opt_program_name);
    n = strlen(buf);

    fputc(IGNOREEOL,fp);
    fprintf(fp,"%s",buf);
    fputc(IGNOREEOL,fp); fputc('\n',fp);

    fputc(IGNOREEOL,fp);
    for(i=0;i<n;++i) fprintf(fp,"-");
	fputc(IGNOREEOL,fp); fputc('\n',fp);
}
		/* ----------------------------------------- */
		/* write options out to the interactive menu */
		/* ----------------------------------------- */

char *
opt_mstring(int i)
{
    static char mstring[180];
    char	flgstr[10];
    char	*s;
    char	*format = "%c %-40s %s";
    char        *lformat = "  --%s %*s%s";

    if( OPT_isundelimtype( optlist[i].type ) )
		return NULL;

    s = optstrval(i);
    if( optlist[i].type == OPT_INTLEVEL ) {
        int f;
        f = OPT_GET_ITH_VALUE(int,i);
        if(f==0)
            strcpy(flgstr,"OFF");
        else if (f==1)
            strcpy(flgstr,"ON");
        else
            sprintf(flgstr,"ON:%d",f);
        s=flgstr;
    } else if( optlist[i].type == OPT_CHAR ) {
        switch (s[0]) {
        case '\0':	sprintf(flgstr,"\'\\0\'"); break;
        case '\n':	sprintf(flgstr,"\'\\n\'"); break;
        case '\t':	sprintf(flgstr,"\'\\t\'"); break;
        case ' ':	sprintf(flgstr,"\'  \'");  break;
        default:	sprintf(flgstr,"%c",s[0]); break;
        }
        s = flgstr;
    } else if( OPT_isflagtype( optlist[i].type ) ) {
        if( *s == '+' || *s == '1')
            strcpy(flgstr,"TRUE");
        if( *s == '-' || *s == '0')
            strcpy(flgstr,"FALSE");
        s=flgstr;
    }

    /* This needs to be fancier, since we now have many combinarions
     * of long/short name'd options
     */
    if (optlist[i].name) {
        sprintf(mstring,format,optlist[i].name,optlist[i].descript,s);
    } else if (optlist[i].longname) {
        sprintf(mstring,lformat,optlist[i].longname,
                38 - strlen(optlist[i].longname),"",s);
    }
            
    return mstring;
}
int
opt_fprint(FILE *fp, int i)
{
    fprintf(fp,"-%c%-17s %c%s\n",optlist[i].name,
            optstrval(i),IGNOREEOL,optlist[i].descript);
    return i;
}

/* opt_usage: Write a usage statement describing registered options.
 * This is what the program writes in response to '--help' on the
 * command line.
 */

#define OPTUSAGEDEBUG 0

void
opt_usage(void)
{
    static char *dformat = " -%c                 %-12s    %s\n";
    static char *lformat=         " -%c, --%-12s %-12s    %s\n";
    static char *uformat =  "                    %-12s    %s\n";
    static char *aformat =         "     --%-12s %-12s    %s\n";
    int  i;
    
    for(i=0; i<opt_nreg; ++i) {
        if(OPTUSAGEDEBUG) fprintf(stderr,"%2d ",i);
        if( OPT_isundelimtype( optlist[i].type ) ) {
            if(OPTUSAGEDEBUG) fprintf(stderr,"u ");
            opt_mess_2(uformat,optstrtyp(i),optlist[i].descript);
        } else {
            if (optlist[i].name == '\0') {
                    if(OPTUSAGEDEBUG) fprintf(stderr,"a ");
                    opt_mess_3(aformat,optlist[i].longname,
                               optstrtyp(i),optlist[i].descript);
            } else {
                if (!ISEMPTYSTRING(optlist[i].longname)) {
                    if(OPTUSAGEDEBUG) fprintf(stderr,"l ");
                    opt_mess_4(lformat,optlist[i].name,optlist[i].longname,
                               optstrtyp(i),optlist[i].descript);
                } else {
                    if(OPTUSAGEDEBUG) fprintf(stderr,"d ");
                    opt_mess_3(dformat,optlist[i].name,
                               optstrtyp(i),optlist[i].descript);
                }
            }
        }
    }
    if (opt_additional_usage_fcn != NULL) {
        /* precede additional usage by -- to indicate that we
         * are finished with the options
         */
        opt_message(" --\n");
        /* return value is ignored */
        (*opt_additional_usage_fcn)();
    }
}

	/* ----------------------- */
	/* Get and process options */
	/* ----------------------- */

	/*
	 * The routines below make use of the ARGVECTOR structure
	 * defined in "ag.h" to process the command line options
	 */
		
	/* -------------------------------------- */
	/* get and process an UN-delimited option */
	/* -------------------------------------- */

int
opt_undelim(ARGVECTOR *ag)
{
    int     i;
    char    *s;
 
    /* Find the first undelimited option that has not been invoked and
     * set that to the argument string.  Note that once an undelimited
     * option has been invoked, it cannot be un-invoked.
     */

    s = strdup( ag_s_advance(ag) );
    for (i=0; i<opt_nreg; ++i) {
        if (optlist[i].invoked) continue;
        if (optlist[i].type==OPT_UNDELIMC ) {
            opt_CSTRING_deprecated();
            strncpy( (char *)optlist[i].value, s , OPT_MAXSTRLEN);
            optlist[i].invoked += 1;
            return 1; /* success */
        }
        if (optlist[i].type==OPT_UNDELIM ||
            optlist[i].type==OPT_UNDELIMV) {
            /** optlist[i].value = strdup(s); <-- this fails! **/
            OPT_SET_ITH_VALUE(char *,i,strdup(s));
            optlist[i].invoked += 1;
            return 1; /* success */
        }
    }
    return 0; /* failed to find an uninvoked option */
}

		/* ---------------------------------- */
		/* get and process a delimited option */
		/* ---------------------------------- */

int
opt_delim(ARGVECTOR *ag)
{
    int      i;
    char     c;
 
    c = ag_c_advance(ag);   /* first character gives name of option */
    i = opt_char_number(c);      /* this is number of option w/ name 'c' */
 
    if(! OPT_isvalidnum(i)) { /* if invalid option name */
        opt_warn_1("%c not a registered option name\n",c);
        return 0;
    }
    return opt_action(i,ag);
}


int
opt_longdelim(ARGVECTOR *ag)
{
    int      i;
    char     c;
    char     *sc,*scptr;

    /* this code parses input options of the form --var=value */
    
    sc = strdup(ag_s(ag));      /* long option name is a string */
    scptr = sc;
    while( !ag_eow(ag) ) {
        c=ag_c_advance(ag);
        if (c == '=') {
            *scptr = '\0';
            break;
        }
        ++scptr;
    }
    
    i  = opt_longname_number(sc);   /* this is number of option w/ name `sc' */
    if(! OPT_isvalidnum(i)) {
        opt_warn_1("%s not a registered option name\n",sc);
        return 0;
    }
    free((char *)sc); 
    return opt_action(i,ag);
}

static void
opt_setvalue(void *v, opt_TYPE o, char *s)
{
  switch(o) {
  case OPT_INT:
    OPT_SETVALUE(int,v, opt_atoi(s));
    break;
  case OPT_UNSINT: /* deprecated */
  case OPT_UINT:
    OPT_SETVALUE(unsigned int, v, opt_atou(s));
    break;
  case OPT_SHORT:
    OPT_SETVALUE(short, v, opt_atoi(s));
    break;
  case OPT_LONG:
    OPT_SETVALUE(long, v, opt_atoi(s));
    break;
  case OPT_USHORT:
    OPT_SETVALUE(unsigned short, v, opt_atou(s));
    break;
  case OPT_ULONG:
    OPT_SETVALUE(unsigned long, v, opt_atou(s));
    break;
  case OPT_FLOAT:
    OPT_SETVALUE(float, v, opt_atof(s));
    break;
  case OPT_DOUBLE:
    OPT_SETVALUE(double, v, opt_atof(s));
    break;
  case OPT_CHAR:
    OPT_SETVALUE(char, v, s[0]);
    break;
  case OPT_UCHAR:
    OPT_SETVALUE(unsigned char, v, s[0]);
    break;
        
  case OPT_STRING:
  case OPT_VSTRING:
    OPT_SETVALUE(char *, v,strdup(s));
    break;
  case OPT_CSTRING:
    opt_CSTRING_deprecated();
    strncpy( (char *)v , s, OPT_MAXSTRLEN );
    break;
        
  default:
    opt_fatal("Invalid type");
  }
}
    
static void
optarray_action(OptArray_t *arr, char *stmp)
{
  char *s = strdup(stmp);
  char *t = s;
  char *u = s;
  char cu = '\0';
  char *v;  /* why is this a char * and not a void *, even though it
             * is meant as a generic pointer?  see the long comment
             * above in the optarray_getstrval() function
             */

  /* Reset the array values */
  if (*(arr->value)) free(*(arr->value));
  *(arr->value)=NULL;
  *(arr->size)=0;

  while (*t) {
    for (u=t; !(*u == '\0' || *u == arr->delim); ++u) {
      /* read until the first delimiter */
      /* t is first character of word */
      /* u is delimiter, so u-1 is last character of t's word */
    }
    cu = *u;
    *u = '\0';  /* now t is null-terminated string */

    /* add new element to array */
    *(arr->size) += 1;
    *(arr->value) = realloc(*(arr->value),
                            *(arr->size)*optsizeof(arr->base_type));
    /* v points to the next element in the array */
    v = (char *)(*(arr->value)) + (*(arr->size)-1)*optsizeof(arr->base_type);
    /* set the value of v according to the string t */
    opt_setvalue(v,arr->base_type,t);
    /* now go to the next word */
    if (cu) t = u+1;
    else    t = u; /* unless that was the last word */
  }
  free(s);
}
                
static int
opt_action(int i, ARGVECTOR *ag)
{
    /* ------------- */
    /* act on option */
    /* ------------- */
    opt_TYPE  o;
    char      c;
    char      *s;
    int       yes,no,maybe,toggle;    /* flag values */

    if (!OPT_isvalidnum(i)) {
        opt_warn_1("opt_action: invalid number %d\n",i);
        return 0; /* failure */
    }

    switch( o=optlist[i].type ) {
    case OPT_INT:
    case OPT_UNSINT: /* deprecated */
    case OPT_UINT:
    case OPT_SHORT:
    case OPT_LONG:
    case OPT_USHORT:
    case OPT_ULONG:
    case OPT_FLOAT:
    case OPT_DOUBLE:
      opt_setvalue(optlist[i].value,o,argnext(ag));
      break;
    case OPT_CHAR:
      OPT_SET_ITH_VALUE(char,i,ag_c_advance(ag));
      break;
    case OPT_UCHAR:
      OPT_SET_ITH_VALUE(unsigned char,i,ag_c_advance(ag));
      break;

    case OPT_FLAG:
    case OPT_NEGFLAG:
    case OPT_ABSFLAG:
    case OPT_ABSNEGFLAG:
        /* define terms:
         * yes: value of flag when  ON
         * no:  value of flag when OFF
         * maybe:  present value of flag
         * toggle: negative of maybe
         */
        maybe  = OPT_GET_ITH_VALUE(int,i);
        toggle = (maybe==OPT_TRUE ? OPT_FALSE : OPT_TRUE);
        if( o==OPT_FLAG || o==OPT_ABSFLAG )
            yes=OPT_TRUE;
        else
            yes=OPT_FALSE;
        no = (yes==OPT_TRUE ? OPT_FALSE : OPT_TRUE);
        c = ag_c(ag);           /* character following 'c' */
        switch(c) {
        case '+':
        case '1':
            OPT_SET_ITH_VALUE(int,i,yes);
            ag_c_advance(ag);       /* eat the '+' */
            break;
        case '-':
        case '0':
            OPT_SET_ITH_VALUE(int,i,no);
            ag_c_advance(ag);       /* eat the '-' */
            break;
        default:
            if( o==OPT_ABSFLAG || o==OPT_ABSNEGFLAG ) {
                OPT_SET_ITH_VALUE(int,i,yes);
            } else {    /* otherwise toggle value */
                OPT_SET_ITH_VALUE(int,i,toggle);
            }
            break;
        }
        break;

    case OPT_INTLEVEL:
      /* default is to increment */
        OPT_SET_ITH_VALUE(int,i,OPT_GET_ITH_VALUE(int,i) + 1);  
        c = ag_c(ag);
        switch(c) {
        case '+':
            /* we've already incremented */
            ag_c_advance(ag);
            break;
        case '-':
            OPT_SET_ITH_VALUE(int,i,0);
            ag_c_advance(ag);
            break;
        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9':
            OPT_SET_ITH_VALUE(int,i,c-'0');
            ag_c_advance(ag);
        default:
            break;
        }
        break;
        
    case OPT_STRING:
    case OPT_VSTRING:
        s = argnext(ag);
        OPT_SET_ITH_VALUE(char *,i,strdup(s));
        /* If this type is really an array, then act accordingly */
        if (optlist[i].array) {
          optarray_action(optlist[i].array,s);
        }
        break;
    case OPT_CSTRING:
        opt_CSTRING_deprecated();
        strncpy( (char *)optlist[i].value , argnext(ag), OPT_MAXSTRLEN );
        break;
        
    case OPT_UNDELIMC:
    case OPT_UNDELIMV:
        opt_fatal("opt_action: can't process UNDELIM type");
        break;

    case OPT_NUL:
        break;
        
    default:
        opt_fatal("opt_delim: don't know this type");
        break;
    }


    /* Indicate that the option was invoked */
    optlist[i].invoked += 1;

    /* Run the hook if its defined */
    if (optlist[i].hook) {
        int retval;
        retval = optlist[i].hook(optlist[i].value);

        if (retval == OPT_ERROR) {
            /* do nothing for now */;
        }
        if (retval == OPT_ABORT && !menuflag)
            exit(opt_exit_number);
        if (retval == OPT_QUIT)
            opt_quit();
        if (retval == OPT_EXIT)
            exit(opt_exit_number);
    }
    
    return 1;                  /* success */
}















